@@ -414,6 +414,8 @@ struct stats {
     uint64_t      extstore_compact_lost; /* items lost because they were locked */
     uint64_t      extstore_compact_rescues; /* items re-written during compaction */
     uint64_t      extstore_compact_skipped; /* unhit items skipped during compaction */
+    uint64_t      extstore_compact_resc_cold; /* items re-written during compaction */
+    uint64_t      extstore_compact_resc_old; /* items re-written during compaction */
 #endif
 #ifdef TLS
     uint64_t      ssl_handshake_errors; /* TLS failures at accept/handshake time */
@@ -457,7 +459,6 @@ struct settings {
     char *inter;
     int verbose;
     rel_time_t oldest_live; /* ignore existing items older than this */
-    uint64_t oldest_cas; /* ignore existing items with CAS values lower than this */
     int evict_to_free;
     char *socketpath;   /* path to unix socket if using local socket */
     char *auth_file;    /* path to user authentication file */
@@ -547,6 +548,7 @@ struct settings {
     bool proxy_uring; /* if the proxy should use io_uring */
     bool proxy_memprofile; /* output detail of lua allocations */
     char *proxy_startfile; /* lua file to run when workers start */
+    char *proxy_startarg; /* string argument to pass to proxy */
     void *proxy_ctx; /* proxy's state context */
 #endif
 #ifdef SOCK_COOKIE_ID
@@ -749,6 +751,11 @@ typedef struct {
     void *proxy_event_thread; // worker threads can also be proxy IO threads
     struct event *proxy_gc_timer; // periodic GC pushing.
     pthread_mutex_t proxy_limit_lock;
+    int proxy_vm_extra_kb;
+    int proxy_vm_last_kb;
+    unsigned int proxy_vm_negative_delta;
+    int proxy_vm_gcrunning;
+    bool proxy_vm_needspoke;
     uint64_t proxy_active_req_limit;
     uint64_t proxy_buffer_memory_limit; // protected by limit_lock
     uint64_t proxy_buffer_memory_used; // protected by limit_lock
@@ -947,7 +954,7 @@ enum delta_result_type do_add_delta(LIBEVENT_THREAD *t, const char *key,
                                     const int64_t delta, char *buf,
                                     uint64_t *cas, const uint32_t hv,
                                     item **it_ret);
-enum store_item_type do_store_item(item *item, int comm, LIBEVENT_THREAD *t, const uint32_t hv, int *nbytes, uint64_t *cas, bool cas_stale);
+enum store_item_type do_store_item(item *item, int comm, LIBEVENT_THREAD *t, const uint32_t hv, int *nbytes, uint64_t *cas, const uint64_t cas_in, bool cas_stale);
 void thread_io_queue_add(LIBEVENT_THREAD *t, int type, void *ctx, io_queue_stack_cb cb);
 void conn_io_queue_setup(conn *c);
 io_queue_t *conn_io_queue_get(conn *c, int type);
@@ -960,6 +967,7 @@ void conn_worker_readd(conn *c);
 extern int daemonize(int nochdir, int noclose);
 
 #define mutex_lock(x) pthread_mutex_lock(x)
+#define mutex_trylock(x) pthread_mutex_trylock(x)
 #define mutex_unlock(x) pthread_mutex_unlock(x)
 
 #include "stats_prefix.h"
@@ -1003,7 +1011,7 @@ item *item_get_locked(const char *key, const size_t nkey, LIBEVENT_THREAD *t, co
 item *item_touch(const char *key, const size_t nkey, uint32_t exptime, LIBEVENT_THREAD *t);
 int   item_link(item *it);
 void  item_remove(item *it);
-int   item_replace(item *it, item *new_it, const uint32_t hv);
+int   item_replace(item *it, item *new_it, const uint32_t hv, const uint64_t cas_in);
 void  item_unlink(item *it);
 
 void item_lock(uint32_t hv);
@@ -1029,7 +1037,7 @@ LIBEVENT_THREAD *get_worker_thread(int id);
 void append_stat(const char *name, ADD_STAT add_stats, conn *c,
                  const char *fmt, ...);
 
-enum store_item_type store_item(item *item, int comm, LIBEVENT_THREAD *t, int *nbytes, uint64_t *cas, bool cas_stale);
+enum store_item_type store_item(item *item, int comm, LIBEVENT_THREAD *t, int *nbytes, uint64_t *cas, const uint64_t cas_in, bool cas_stale);
 
 /* Protocol related code */
 void out_string(conn *c, const char *str);
@@ -1057,7 +1065,7 @@ void conn_set_state(conn *c, enum conn_states state);
 void out_of_memory(conn *c, char *ascii_error);
 void out_errstring(conn *c, const char *str);
 void write_and_free(conn *c, char *buf, int bytes);
-void server_stats(ADD_STAT add_stats, conn *c);
+void server_stats(ADD_STAT add_stats, void *c);
 void append_stats(const char *key, const uint16_t klen,
                   const char *val, const uint32_t vlen,
                   const void *cookie);
