@@ -328,7 +328,7 @@ static void complete_incr_bin(conn *c, char *extbuf) {
                 memcpy(ITEM_data(it) + res, "\r\n", 2);
                 c->thread->cur_sfd = c->sfd; // for store_item logging.
 
-                if (store_item(it, NREAD_ADD, c->thread, NULL, &cas, CAS_NO_STALE)) {
+                if (store_item(it, NREAD_ADD, c->thread, NULL, &cas, (settings.use_cas) ? get_cas_id() : 0, CAS_NO_STALE)) {
                     c->cas = cas;
                     write_bin_response(c, &rsp->message.body, 0, 0, sizeof(rsp->message.body.value));
                 } else {
@@ -386,7 +386,7 @@ static void complete_update_bin(conn *c) {
 
     uint64_t cas = 0;
     c->thread->cur_sfd = c->sfd; // for store_item logging.
-    ret = store_item(it, c->cmd, c->thread, NULL, &cas, CAS_NO_STALE);
+    ret = store_item(it, c->cmd, c->thread, NULL, &cas, (settings.use_cas) ? get_cas_id() : 0, CAS_NO_STALE);
     c->cas = cas;
 
 #ifdef ENABLE_DTRACE
@@ -737,7 +737,7 @@ static void process_bin_sasl_auth(conn *c) {
 
     assert(c->binary_header.request.extlen == 0);
 
-    int nkey = c->binary_header.request.keylen;
+    uint16_t nkey = c->binary_header.request.keylen;
     int vlen = c->binary_header.request.bodylen - nkey;
 
     if (nkey > MAX_SASL_MECH_LEN) {
@@ -776,7 +776,7 @@ static void process_bin_complete_sasl_auth(conn *c) {
     assert(c->item);
     init_sasl_conn(c);
 
-    int nkey = c->binary_header.request.keylen;
+    uint16_t nkey = c->binary_header.request.keylen;
     int vlen = c->binary_header.request.bodylen - nkey;
 
     if (nkey > ((item*) c->item)->nkey) {
@@ -1082,7 +1082,7 @@ static void dispatch_bin_command(conn *c, char *extbuf) {
 
 static void process_bin_update(conn *c, char *extbuf) {
     char *key;
-    int nkey;
+    uint16_t nkey;
     int vlen;
     item *it;
     protocol_binary_request_set* req = (void *)extbuf;
@@ -1191,7 +1191,7 @@ static void process_bin_update(conn *c, char *extbuf) {
 
 static void process_bin_append_prepend(conn *c) {
     char *key;
-    int nkey;
+    uint16_t nkey;
     int vlen;
     item *it;
 
@@ -1268,17 +1268,12 @@ static void process_bin_flush(conn *c, char *extbuf) {
     }
 
     if (exptime > 0) {
-        new_oldest = realtime(exptime);
+        new_oldest = realtime(exptime) - 1;
     } else {
-        new_oldest = current_time;
-    }
-    if (settings.use_cas) {
-        settings.oldest_live = new_oldest - 1;
-        if (settings.oldest_live <= current_time)
-            settings.oldest_cas = get_cas_id();
-    } else {
-        settings.oldest_live = new_oldest;
+        new_oldest = current_time - 1;
     }
+    settings.oldest_live = new_oldest;
+    item_flush_expired();
 
     pthread_mutex_lock(&c->thread->stats.mutex);
     c->thread->stats.flush_cmds++;
