@@ -310,10 +310,9 @@ int mcplib_await_return(io_pending_proxy_t *p) {
     // lose our internal mcpres reference regardless.
     // also tag the elapsed time into the response.
     if (p->mcpres_ref) {
-        struct timeval end;
-        gettimeofday(&end, NULL);
-        p->client_resp->elapsed = (end.tv_sec - p->client_resp->start.tv_sec) * 1000000 +
-            (end.tv_usec - p->client_resp->start.tv_usec);
+        // NOTE: this is redundant but the code is going away soon. not worth
+        // testing changing it.
+        mcp_resp_set_elapsed(p->client_resp);
 
         // instructed to generate log_req entries for each failed request,
         // this is useful to do here as these can be asynchronous.
@@ -335,7 +334,7 @@ int mcplib_await_return(io_pending_proxy_t *p) {
                     lua_pop(L, 1);
                 }
 
-                logger_log(l, LOGGER_PROXY_REQ, NULL, rq->pr.request, rq->pr.reqlen, rs->elapsed, rs->resp.type, rs->resp.code, rs->status, detail, dlen, rs->be_name, rs->be_port);
+                logger_log(l, LOGGER_PROXY_REQ, NULL, rq->pr.request, rq->pr.reqlen, rs->elapsed, rs->resp.type, rs->resp.code, rs->status, aw->rctx->conn_fd, detail, dlen, rs->be_name, rs->be_port);
             }
         }
 
