@@ -525,9 +525,16 @@ static void *worker_libevent(void *arg) {
     }
 
     register_thread_initialized();
-
+#ifdef PROXY
+    while (!event_base_got_exit(me->base)) {
+        event_base_loop(me->base, EVLOOP_ONCE);
+        if (me->proxy_ctx) {
+            proxy_gc_poke(me);
+        }
+    }
+#else
     event_base_loop(me->base, 0);
-
+#endif
     // same mechanism used to watch for all threads exiting.
     register_thread_initialized();
 
@@ -677,7 +684,7 @@ static void thread_libevent_process(evutil_socket_t fd, short which, void *arg)
 #define THR_NAME_MAXLEN 16
 void thread_setname(pthread_t thread, const char *name) {
 assert(strlen(name) < THR_NAME_MAXLEN);
-#if defined(__linux__)
+#if defined(__linux__) && defined(HAVE_PTHREAD_SETNAME_NP)
 pthread_setname_np(thread, name);
 #endif
 }
@@ -905,20 +912,6 @@ item *item_touch(const char *key, size_t nkey, uint32_t exptime, LIBEVENT_THREAD
     return it;
 }
 
-/*
- * Links an item into the LRU and hashtable.
- */
-int item_link(item *item) {
-    int ret;
-    uint32_t hv;
-
-    hv = hash(ITEM_key(item), item->nkey);
-    item_lock(hv);
-    ret = do_item_link(item, hv);
-    item_unlock(hv);
-    return ret;
-}
-
 /*
  * Decrements the reference count on an item and adds it to the freelist if
  * needed.
@@ -937,8 +930,8 @@ void item_remove(item *item) {
  * Unprotected by a mutex lock since the core server does not require
  * it to be thread-safe.
  */
-int item_replace(item *old_it, item *new_it, const uint32_t hv) {
-    return do_item_replace(old_it, new_it, hv);
+int item_replace(item *old_it, item *new_it, const uint32_t hv, const uint64_t cas_in) {
+    return do_item_replace(old_it, new_it, hv, cas_in);
 }
 
 /*
@@ -972,13 +965,13 @@ enum delta_result_type add_delta(LIBEVENT_THREAD *t, const char *key,
 /*
  * Stores an item in the cache (high level, obeys set/add/replace semantics)
  */
-enum store_item_type store_item(item *item, int comm, LIBEVENT_THREAD *t, int *nbytes, uint64_t *cas, bool cas_stale) {
+enum store_item_type store_item(item *item, int comm, LIBEVENT_THREAD *t, int *nbytes, uint64_t *cas, const uint64_t cas_in, bool cas_stale) {
     enum store_item_type ret;
     uint32_t hv;
 
     hv = hash(ITEM_key(item), item->nkey);
     item_lock(hv);
-    ret = do_store_item(item, comm, t, hv, nbytes, cas, cas_stale);
+    ret = do_store_item(item, comm, t, hv, nbytes, cas, cas_in, cas_stale);
     item_unlock(hv);
     return ret;
 }
